try:
    from ._version import __version__
except ImportError:
    try:
        from setuptools_scm import get_version

        __version__ = get_version()
    except ImportError:
        __version__ = "unknown"
