from __future__ import annotations

import defcon
from fontParts.base import BaseFeatures
from fontParts.fontshell.base import RBaseObject


class RFeatures(RBaseObject, BaseFeatures):
    wrapClass = defcon.Features

    def _get_text(self) -> str | None:
        features = self.naked()
        return features.text

    def _set_text(self, value: str | None) -> None:
        features = self.naked()
        features.text = value
